"""Default values for KLayout DRC."""

from __future__ import annotations

from pathlib import Path

DEFAULT_GDSFILE = Path("layout.gds")
DEFAULT_RESULTSFILE = Path("drc_results.lyrdb")
DEFAULT_VERBOSE = True
